from typing import Dict, Optional, cast

from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse

from django_backblaze_b2.options import BackblazeB2StorageOptions
from django_backblaze_b2.storage import BackblazeB2Storage


class PublicStorage(BackblazeB2Storage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bucketInfo: Optional[bool] = None

    def _isPublicBucket(self) -> bool:
        if self._bucketInfo is None:
            self._bucketInfo = self.bucket.as_dict().get("bucketType") == "allPublic"
        return self._bucketInfo

    def _getFileUrl(self, name: str) -> str:
        if self._isPublicBucket():
            return self.getBackblazeUrl(name)
        return reverse("django_b2_storage:b2-public", args=[name])

    def _getDjangoSettingsOptions(self, kwargOpts: Dict) -> BackblazeB2StorageOptions:
        _validateKwargOpts(kwargOpts)
        options = super()._getDjangoSettingsOptions(kwargOpts)
        _adjustOptions(options, specificBucket="public")
        return options


class LoggedInStorage(BackblazeB2Storage):
    def _getFileUrl(self, name: str) -> str:
        return reverse("django_b2_storage:b2-logged-in", args=[name])

    def _getDjangoSettingsOptions(self, kwargOpts: Dict) -> BackblazeB2StorageOptions:
        _validateKwargOpts(kwargOpts)
        options = super()._getDjangoSettingsOptions(kwargOpts)
        _adjustOptions(options, specificBucket="loggedIn")
        return options


class StaffStorage(BackblazeB2Storage):
    def _getFileUrl(self, name: str) -> str:
        return reverse("django_b2_storage:b2-staff", args=[name])

    def _getDjangoSettingsOptions(self, kwargOpts: Dict) -> BackblazeB2StorageOptions:
        _validateKwargOpts(kwargOpts)
        options = super()._getDjangoSettingsOptions(kwargOpts)
        _adjustOptions(options, specificBucket="staff")
        return options


def _validateKwargOpts(kwargOpts) -> None:
    if "bucket" in kwargOpts:
        raise ImproperlyConfigured("May not specify 'bucket' in proxied storage class")
    if "application_key_id" in kwargOpts or "application_key" in kwargOpts or "realm" in kwargOpts:
        raise ImproperlyConfigured("May not specify auth credentials in proxied storage class")


def _adjustOptions(options: BackblazeB2StorageOptions, specificBucket: str) -> None:
    proxiedStorageBucketName = options.get("specificBucketNames", {}).get(specificBucket)
    if proxiedStorageBucketName:
        options["bucket"] = cast(str, proxiedStorageBucketName)



from datetime import datetime
from logging import getLogger
from typing import IO, Any, Dict, List, Optional, Tuple

from b2sdk.v1 import B2Api, Bucket, InMemoryAccountInfo
from b2sdk.v1.exception import FileNotPresent, NonExistentBucket
from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import File
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible

from django_backblaze_b2.b2_file import B2File
from django_backblaze_b2.b2_filemeta_shim import FileMetaShim
from django_backblaze_b2.options import BackblazeB2StorageOptions, getDefaultB2StorageOptions

logger = getLogger("django-backblaze-b2")


@deconstructible
class BackblazeB2Storage(Storage):
    """Storage class which fulfills the Django Storage contract through b2 apis"""

    def __init__(self, **kwargs):
        opts = self._getDjangoSettingsOptions(kwargs.get("opts", {}))
        opts.update(kwargs.get("opts", {}))

        self._bucketName = opts["bucket"]
        self._defaultFileInfo = opts["defaultFileInfo"]
        self._authInfo = dict(
            [(k, v) for k, v in opts.items() if k in ["realm", "application_key_id", "application_key"]]
        )
        self._allowFileOverwrites = opts["allowFileOverwrites"]

        logger.info(f"{self.__class__.__name__} instantiated to use bucket {self._bucketName}")
        if opts["authorizeOnInit"]:
            logger.debug(f"{self.__class__.__name__} authorizing")
            self.b2Api
            if opts["validateOnInit"]:
                self._getOrCreateBucket(opts["nonExistentBucketDetails"])

    def _getDjangoSettingsOptions(self, kwargOpts: Dict) -> BackblazeB2StorageOptions:
        """Setting terminology taken from:
        https://b2-sdk-python.readthedocs.io/en/master/glossary.html#term-application-key-ID
        """
        from django.conf import settings

        if not hasattr(settings, "BACKBLAZE_CONFIG"):
            raise ImproperlyConfigured("add BACKBLAZE_CONFIG dict to django settings")
        if "application_key_id" not in settings.BACKBLAZE_CONFIG or "application_key" not in settings.BACKBLAZE_CONFIG:
            raise ImproperlyConfigured(
                "At minimium BACKBLAZE_CONFIG must contain auth 'application_key' and 'application_key_id'"
                f"\nfound: {settings.BACKBLAZE_CONFIG}"
            )
        opts = getDefaultB2StorageOptions()
        opts.update(settings.BACKBLAZE_CONFIG)  # type: ignore
        return opts

    @property
    def b2Api(self) -> B2Api:
        if not hasattr(self, "_b2Api"):
            self._accountInfo = InMemoryAccountInfo()
            self._b2Api = B2Api(self._accountInfo)
            self._b2Api.authorize_account(**self._authInfo)
        return self._b2Api

    @property
    def bucket(self) -> Bucket:
        if not hasattr(self, "_bucket"):
            self._getOrCreateBucket()
        return self._bucket

    def _getOrCreateBucket(self, newBucketDetails=None) -> None:
        try:
            self._bucket = self.b2Api.get_bucket_by_name(self._bucketName)
        except NonExistentBucket as e:
            if newBucketDetails is not None:
                logger.debug(f"Bucket {self._bucketName} not found. Creating with details: {newBucketDetails}")
                if "bucket_type" not in newBucketDetails:
                    newBucketDetails["bucket_type"] = "allPrivate"
                self._bucket = self.b2Api.create_bucket(name=self._bucketName, **newBucketDetails)
            else:
                raise e
        logger.debug(f"Connected to bucket {self._bucket.as_dict()}")

    def _open(self, name: str, mode: str) -> File:
        return B2File(
            name=name, bucket=self.bucket, fileInfos=self._defaultFileInfo, mode=mode, sizeProvider=self.size,
        )

    def _save(self, name: str, content: IO[Any]) -> str:
        """
        Save and retrieve the filename.
        If the file exists it will make another version of that file.
        """
        return B2File(
            name=name, bucket=self.bucket, fileInfos=self._defaultFileInfo, mode="w", sizeProvider=self.size,
        ).saveAndRetrieveFile(content)

    def path(self, name: str) -> str:
        return name

    def delete(self, name: str) -> None:
        try:
            fileId = FileMetaShim(self.b2Api, self.bucket, name).id
            logger.debug(f"Deleting file {name} id=({fileId})")
            self.b2Api.delete_file_version(file_id=fileId, file_name=name)
        except FileNotPresent:
            logger.debug("Not found")

    def exists(self, name: str) -> bool:
        return FileMetaShim(self.b2Api, self.bucket, name).exists

    def size(self, name: str) -> int:
        return FileMetaShim(self.b2Api, self.bucket, name).contentLength

    def url(self, name: Optional[str]) -> str:
        if not name:
            raise Exception("Name must be defined")
        return self._getFileUrl(name)

    def _getFileUrl(self, name: str) -> str:
        return self.getBackblazeUrl(name)

    def getBackblazeUrl(self, filename: str) -> str:
        return self.b2Api.get_download_url_for_file_name(bucket_name=self._bucketName, file_name=filename)

    def get_available_name(self, name: str, max_length: Optional[int] = None) -> str:
        if self._allowFileOverwrites:
            return name
        return super().get_available_name(name, max_length)

    def listdir(self, path: str) -> Tuple[List[str], List[str]]:
        """
        List the contents of the specified path. Return a 2-tuple of lists:
        the first item being directories, the second item being files.
        """
        raise NotImplementedError("subclasses of Storage must provide a listdir() method")

    def get_accessed_time(self, name: str) -> datetime:
        """
        Return the last accessed time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError("subclasses of Storage must provide a get_accessed_time() method")

    def get_created_time(self, name: str) -> datetime:
        """
        Return the creation time (as a datetime) of the file specified by name.
        The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError("subclasses of Storage must provide a get_created_time() method")

    def get_modified_time(self, name: str) -> datetime:
        """
        Return the last modified time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError("subclasses of Storage must provide a get_modified_time() method")



from io import BytesIO
from logging import getLogger
from typing import IO, Any, Callable, Dict, Optional

from b2sdk.v1 import Bucket, DownloadDestBytes
from django.core.files.base import File

logger = getLogger("django-backblaze-b2")


class B2File(File):
    """Read/Write as lazy as possible"""

    def __init__(
        self, name: str, bucket: Bucket, sizeProvider: Callable[[str], int], fileInfos: Dict[str, str], mode: str,
    ):
        self.name: str = name
        self._bucket: Bucket = bucket
        self._sizeProvider = sizeProvider
        self._fileInfos = fileInfos
        self._mode: str = mode
        self._hasUnwrittenData: bool = False
        self._contents: Optional[IO] = None

    # https://github.com/python/mypy/issues/4125 -- kinda makes you wonder why we like mypy?
    @property  # type: ignore
    def file(self) -> IO[Any]:  # type: ignore
        if self._contents is None:
            data = self._readFileContents()
            self._contents = BytesIO(data)
        return self._contents

    # https://github.com/python/mypy/issues/1465
    @file.setter  # type: ignore
    def file(self, value: IO[Any]) -> None:
        self._contents = value

    def _readFileContents(self) -> bytes:
        download_dest = DownloadDestBytes()
        self._bucket.download_file_by_name(file_name=self.name, download_dest=download_dest)
        return download_dest.get_bytes_written()

    @property
    def size(self) -> int:
        if not hasattr(self, "_size"):
            self._size = self._sizeProvider(self.name)
        return self._size

    def read(self, num_bytes: Optional[int] = None) -> bytes:
        return self.file.read(num_bytes if isinstance(num_bytes, int) else -1)

    def write(self, content: bytes) -> int:
        if "w" not in self._mode:
            raise AttributeError("File was not opened for write access.")
        self.file = BytesIO(content)

        self._hasUnwrittenData = True
        return len(content)

    def close(self) -> None:
        if self._hasUnwrittenData:
            self.saveAndRetrieveFile(self.file)
        self.file.close()

    def saveAndRetrieveFile(self, content: IO[Any]) -> str:
        """
        Save and retrieve the filename.
        If the file exists it will make another version of that file.
        """
        logger.debug(f"Saving {self.name} to b2 bucket ({self._bucket.get_id()})")
        self._bucket.upload_bytes(
            data_bytes=content.read(), file_name=self.name, file_infos=self._fileInfos,
        )
        self._hasUnwrittenData = False
        return self.name



from typing import Any, Dict, Literal, Optional, TypedDict, Union


class ProxiedBucketNames(TypedDict):
    public: Optional[str]
    loggedIn: Optional[str]
    staff: Optional[str]


class BackblazeB2StorageOptions(TypedDict):
    """Configuration options. Literals indicate defaults"""

    realm: Literal["production"]
    application_key_id: str
    application_key: str
    bucket: str
    authorizeOnInit: bool
    validateOnInit: bool
    allowFileOverwrites: bool
    # see: https://b2-sdk-python.readthedocs.io/en/master/api/api.html#b2sdk.v1.B2Api.create_bucket
    nonExistentBucketDetails: Optional[Dict[str, Union[str, Dict[str, Any]]]]
    defaultFileInfo: Dict[str, Any]
    specificBucketNames: ProxiedBucketNames


def getDefaultB2StorageOptions() -> BackblazeB2StorageOptions:
    return {
        "realm": "production",
        "application_key_id": "---",
        "application_key": "---",
        "bucket": "django",
        "authorizeOnInit": True,
        "validateOnInit": True,
        "allowFileOverwrites": False,
        "nonExistentBucketDetails": None,
        "defaultFileInfo": {},
        "specificBucketNames": {"public": None, "loggedIn": None, "staff": None},
    }



from logging import getLogger

from b2sdk.exception import B2Error, interpret_b2_error
from b2sdk.v1 import B2Api, Bucket
from requests import HTTPError, Response
from requests.models import CaseInsensitiveDict

logger = getLogger("django-backblaze-b2")


class FileMetaShim:
    """
    Shim until you can get file info by name:
    https://github.com/Backblaze/b2-sdk-python/issues/143

    Internally, does a HEAD on the file's url.
    Caches the result in-memory for the lifetime of this object's existence
    """

    def __init__(self, b2Api: B2Api, bucketInstance: Bucket, filename: str) -> None:
        self._b2Api = b2Api
        self._bucket = bucketInstance
        self._filename = filename

    @property
    def id(self) -> str:
        return self.as_dict()["x-bz-file-id"]

    @property
    def contentLength(self) -> int:
        return self.as_dict()["Content-Length"]

    @property
    def exists(self) -> bool:
        try:
            self.as_dict()  # obtain meta via Http
            return True
        except HTTPError as e:
            if e.response.status_code == 404:
                return False
            try:
                raise interpret_b2_error(
                    e.response.status_code,
                    e.response.json().get("code"),
                    e.response.json().get("message"),
                    e.response.headers,
                )
            except B2Error:
                raise
            except Exception:
                logger.exception("Could not interpret b2 error")
                raise Exception(f"Status: {e.response.status_code}, Content: {e.response.content}")

    def as_dict(self) -> CaseInsensitiveDict:
        if not hasattr(self, "_meta"):
            logger.debug(f"HEAD bucket={self._bucket.name} file={self._filename}")
            head_attempts = 4
            for attempt in range(0, head_attempts):
                try:
                    self._meta = self._get_head_response().headers
                    return self._meta
                except HTTPError as e:
                    if attempt == 3 or e.response.status_code == 404:
                        raise
                    if not self._b2Api.authorize_automatically():
                        raise

            # try one last time
            self._meta = self._get_head_response().headers
        return self._meta

    def _get_head_response(self) -> Response:
        downloadUrl = self._b2Api.session.get_download_url_by_name(self._bucket.name, self._filename)
        downloadAuthorization = self._bucket.get_download_authorization(self._filename, 30)
        response = self._b2Api.raw_api.b2_http.session.head(
            downloadUrl, headers={"Authorization": downloadAuthorization}
        )
        response.raise_for_status()
        return response



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_backblaze_b2']

package_data = \
{'': ['*']}

install_requires = \
['b2sdk>=1.1.4,<2.0.0', 'django>=3.0']

setup_kwargs = {
    'name': 'django-backblaze-b2',
    'version': '1.1.2',
    'description': 'A Django app to use backblaze b2 as storage.',
    'long_description': '# django-backblaze-b2\n\n[![pypi version](https://img.shields.io/pypi/v/django-backblaze-b2)](https://pypi.org/project/django-backblaze-b2/)\n[![python version](https://img.shields.io/pypi/pyversions/django-backblaze-b2)](https://pypi.org/project/django-backblaze-b2/)\n[![django version](https://img.shields.io/pypi/djversions/django-backblaze-b2)](https://pypi.org/project/django-backblaze-b2/)\n\nA storage backend for Django that uses [Backblaze\'s B2 APIs](https://www.backblaze.com/b2/cloud-storage.html).\n\nImplementation wraps [Official Python SDK](https://github.com/Backblaze/b2-sdk-python)\n\n## How to use\n\n1. Install from this repo, or install from PyPi: `pip install django-backblaze-b2`\nAs tested, requires python 3.6 or greater but solely due to type annotations. PRs welcome :)\n1. Configure your django `settings`. The absolute minimum config would be:\n```python\nBACKBLAZE_CONFIG = {\n    "application_key_id": os.getenv("BACKBLAZE_KEY_ID"), # however you want to securely retrieve these values\n    "application_key": os.getenv("BACKBLAZE_KEY"),\n}\n```\n\nTheoretically you may now refer to the base storage class as a storage class.  \ne.g.\n```python\nfrom django_backblaze_b2 import BackblazeB2Storage\n\nclass MyModel(models.Model):\n    fileField = models.FileField(\n        upload_to="uploads",\n        storage=BackblazeB2Storage\n    )\n```\n\n### Public/Logged-In/Private storage\n\n1. Add `django_backblaze_b2` to your `INSTALLED_APPS`\n1. Add the urls to your `urlpatterns` in the root `urls.py`:\n```python\n    urlpatterns = [\n        ...\n        path(\'\', include(\'django_backblaze_b2.urls\')),\n    ]\n```\n\n### Configurations \n\nYou may want to use your own bucket name, or set further configuration such as lazy authorization/validation, or specifying file metadata.  \nRefer to [the options](./django_backblaze_b2/options.py) for all options.  \nYou can modify the settings dict, but additionally override any setting with the `opts` keyword argument to the storage classes.\n\nTo specify different buckets to use for your public, logged-in, staff storage, you can set the \n`specificBucketNames` attribute of the settings dict.\n\n## Why\n\nThere are several Django storage packages out there already which support B2, but none met my needs. These are:\n\n* [django-storages](https://github.com/jschneier/django-storages)\n    * Large community engagement ✅\n    * Well-tested ✅\n    * [Second-class support](https://github.com/jschneier/django-storages/issues/765) via [Apache Libcloud](https://github.com/apache/libcloud) ❌\n    * Disconnect in configuration and actual use ❌\n    * PR list with low turnaround ❌\n* [django-b2](https://github.com/pyutil/django-b2)\n    * Similar aim to this project, around official backblaze SDK ✅\n    * Mixed goals (storage, scripts) ❌\n    * Tests?? ❌\n* [django-backblazeb2-storage](https://github.com/royendgel/django-backblazeb2-storage)\n    * Simple configuration ✅\n    * Not based around python SDK (potentially harder to keep up with version changes) ❌\n    * Tests?? ❌\n\n### S3 Compatible API\n\nBackblazed can be used with an [S3-compatible API](https://www.backblaze.com/b2/docs/s3_compatible_api.html)\nThis is great, but most packages use an older version of the S3 Api (v2). Backblaze uses v4.\n\n### What this package offers\n\n* Type Annotations\n* Tested\n* No hacks required to get up and running around API deficiencies (any hacks are not exposed in API)\n* Support for public/private files, restricted via Django user permissions\n\n## How it works\n\n* A simple implementation of the `django.core.files.storage.Storage` class provides handling for storage behaviour within your Django application\n* Three url routes are appended to the root of your application:  \n    1. `/b2/`\n    2. `/b2l/`\n    3. `/b2s/`\nThese routes act as a proxy/intermediary between the requester and backblaze b2 apis. The public `/b2/` allows exposing files from a private bucket, and the logged-in and staff routes will perform the known validations of a django app to prevent unauthorized access.\n\n### Gotchas\n\n* The original filename + any upload paths is stored in the database. Thus your column name must be of sufficient length to hold that (unchanged behaviour from `FileSystemStorage`)\n*  When retrieving files from the `PublicStorage`, `LoggedInStorage` or `StaffStorage`, you may not override the `"bucket"` or authorization options, or else when the app proxies the file download, it will be unable to retrieve the file from the respective bucket.\n* Simply using `LoggedInStorage` or `StaffStorage` is not enough to protect your files if your bucket is not public. If any individual gains access to the file ids/urls for these files, there is no authentication around them. It is up to the implementer to ensure the security of their application.\n* Once the file is uploaded, and someone obtains a file url (e.g. http://djangodomain.com/b2l/uploads/image.png), the model will no longer be checked for the file. This means that if you share the bucket between multiple use-cases, you could in theory find finds that don\'t belong to your django app, or similarly if you delete/change your models, the files could still be downloaded. Consider using an app like [django-cleanup](https://github.com/un1t/django-cleanup) if this is important to you\n\n## Contributing\n\nContributions welcome!\n\n* Please ensure test coverage does not decrease in a meaningful way.\n* Ensure formatting is compliant (`make lint`)\n* Use [conventional commits](https://www.conventionalcommits.org/en/v1.0.0/)\n\n## Setting up for development\n\n### Requires\n\n* python\n* GNU Make\n* (optional) pyenv - align local version\n* (optional) docker - run sample app\n\n### Running\n\n1. `make setup`\n\n* You can run django with `make run-django` to test django app.\n* You can run tests with `make test`\n* You can view test coverage with `make test-coverage`, then see in the terminal, \nopen `test/htmlcov/index.html`\nor use `cov.xml` in your favourite IDE like VSCode\n\n### Releasing\n\n1. `make publish-to-pypi`\n\n### Cleanup\n\n1. `make cleanup`\n',
    'author': 'Étienne H',
    'author_email': 'django_backblaze_b2@internet-e-mail.com',
    'maintainer': 'Étienne H',
    'maintainer_email': 'django_backblaze_b2@internet-e-mail.com',
    'url': 'https://github.com/ehossack/django-backblaze-b2/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
